/* GTK+ interface
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 */

#ifdef WINDOWS
// dumb hack for mingw
typedef long uid_t;
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include "spc.h"
#include "openspc.h"

// global vars
static struct spc_format *spc = NULL;
static GtkListStore *sample_list_store;
static GtkTreeViewColumn *sample_list_column[5];
static GtkTreeModel *sample_list_store_filtered;
static GtkWidget *label_filename, *label_game_title, *label_song_title, *label_artist_name, *label_song_length;
static GtkWidget *frame_filters, *button_play, *spinner[10];
static int chart_filters[10] = { /* min */ 0, 0, 0, 0, 0, /* max */ 0x100, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF};

// helper function to create a label and add it to a grid
static void add_label(GtkWidget **label, char *text, GtkWidget *grid, int left, int top, int width, int height, gboolean is_selectable)
{
	*label = gtk_label_new(text);
	gtk_misc_set_alignment(GTK_MISC(*label), 0.0, 0.5);
	gtk_label_set_selectable(GTK_LABEL(*label), is_selectable);
	gtk_grid_attach(GTK_GRID(grid), *label, left, top, width, height);
}

// helper functions to clean up id666 data
static char *proper_song_length()
{
	char tempseconds[4];
	memcpy(tempseconds, spc->id666_seconds_before_fadeout, 3); tempseconds[3] = '\0';

	int seconds = strtol(tempseconds, NULL, 10);
	int minutes = seconds / 60;
	seconds = seconds % 60;

	char *tempstr = malloc(10);
	snprintf(tempstr, 10, "%d:%02d", minutes, seconds); tempstr[9] = '\0';

	return tempstr;
}

static char *proper_game_title()
{
	char *tempstr = malloc(33);
	snprintf(tempstr, 32, "%s", spc->id666_game_title); tempstr[32] = '\0';
	return tempstr;
}

static char *proper_song_title()
{
	char *tempstr = malloc(33);
	snprintf(tempstr, 32, "%s", spc->id666_song_title); tempstr[32] = '\0';
	return tempstr;
}

static char *proper_artist_name()
{
	char *tempstr = malloc(33);
	snprintf(tempstr, 32, "%s", spc->id666_artist_name); tempstr[32] = '\0';
	return tempstr;
}

// the "About" dialog window
static void about_dialog(GtkWidget *parent_window, gpointer data)
{
	GtkWidget *dialog;

	dialog = gtk_message_dialog_new(GTK_WINDOW(parent_window), GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_INFO, GTK_BUTTONS_CLOSE, "Sample Table Tool\nVersion %d\nBy thedopefish\n\nThis is a thing I made for DSO.", VERSION_NUMBER);

	gtk_dialog_run(GTK_DIALOG(dialog));

	gtk_widget_destroy(dialog);
}

// the "Open SPC" dialog window
static void open_spc_dialog(GtkWidget *parent_window, gpointer data)
{
	GtkWidget *dialog;
	GtkFileFilter *filter;
	GtkFileChooserAction action = GTK_FILE_CHOOSER_ACTION_OPEN;
	gint res;
	GtkTreeIter iter;

	dialog = gtk_file_chooser_dialog_new("Open SPC File", GTK_WINDOW(parent_window), action, "_Cancel", GTK_RESPONSE_CANCEL, "_Open", GTK_RESPONSE_ACCEPT, NULL);

	filter = gtk_file_filter_new();
	gtk_file_filter_set_name(filter, "SPC Files");
	gtk_file_filter_add_pattern(filter, "*.spc");
	gtk_file_filter_add_pattern(filter, "*.SPC");
	gtk_file_chooser_set_filter(GTK_FILE_CHOOSER(dialog), filter);

	res = gtk_dialog_run(GTK_DIALOG(dialog));
	if(res == GTK_RESPONSE_ACCEPT) {
		char *filename;
		GtkFileChooser *chooser = GTK_FILE_CHOOSER(dialog);
		filename = gtk_file_chooser_get_filename(chooser);

		spc = read_spc_from_file(filename);
		if(spc != NULL) {
			// display the ID666 info on the UI
			gtk_label_set_text(GTK_LABEL(label_filename), filename);
			if(spc->has_id666) {
				char *temp = proper_game_title();
				gtk_label_set_text(GTK_LABEL(label_game_title), temp);
				free(temp); temp = proper_song_title();
				gtk_label_set_text(GTK_LABEL(label_song_title), temp);
				free(temp); temp = proper_artist_name();
				gtk_label_set_text(GTK_LABEL(label_artist_name), temp);
				free(temp); temp = proper_song_length();
				gtk_label_set_text(GTK_LABEL(label_song_length), temp);
				free(temp);
			} else {
				gtk_label_set_text(GTK_LABEL(label_game_title), "[No ID666 Info]");
				gtk_label_set_text(GTK_LABEL(label_song_title), "[No ID666 Info]");
				gtk_label_set_text(GTK_LABEL(label_artist_name), "[No ID666 Info]");
				gtk_label_set_text(GTK_LABEL(label_song_length), "[No ID666 Info]");
			}
			gtk_widget_show(button_play);

			// parse interesting data from the sample table
			int num_samples = 0;
			struct sample_chart **chart = NULL;
			generate_sample_chart(spc, &num_samples, &chart);
			if(chart != NULL && num_samples > 0) {
				// nuke the existing sample table (if any) to make room for the new one
				gtk_list_store_clear(sample_list_store);

				struct sample_chart *c = (struct sample_chart*) chart;
				char tempstr[5][20];
				for(int i = 0; i < num_samples; i++) {
					snprintf(tempstr[0], 20, "%X", c->sample_number); tempstr[0][19] = '\0';
					snprintf(tempstr[1], 20, "%X", c->sample_address); tempstr[1][19] = '\0';
					snprintf(tempstr[2], 20, "%X", c->loop_address); tempstr[2][19] = '\0';
					snprintf(tempstr[3], 20, "%X", c->sample_length); tempstr[3][19] = '\0';
					snprintf(tempstr[4], 20, "%X", c->loop_length); tempstr[4][19] = '\0';

					gtk_list_store_append(sample_list_store, &iter);
					gtk_list_store_set(sample_list_store, &iter, 0, tempstr[0], 1, tempstr[1], 2, tempstr[2], 3, tempstr[3], 4, tempstr[4], -1);

					c++;
				}

				free(chart);
			}
		} else {
			gtk_widget_hide(button_play);
			gtk_label_set_text(GTK_LABEL(label_game_title), "");
			gtk_label_set_text(GTK_LABEL(label_song_title), "");
			gtk_label_set_text(GTK_LABEL(label_artist_name), "");
			gtk_label_set_text(GTK_LABEL(label_song_length), "");
		}

		g_free(filename);
	}

	gtk_widget_destroy(dialog);
}

// handler for clicking on the "Show/Hide Filters" toggle
static void filters_click(GtkWidget *widget, gpointer data)
{
	GtkButton *button = GTK_BUTTON(widget);
	if(strcmp("Show Filters", gtk_button_get_label(button)) == 0) {
		gtk_button_set_label(button, "Hide Filters");
		gtk_widget_show(frame_filters);
	} else {
		gtk_button_set_label(button, "Show Filters");
		gtk_widget_hide(frame_filters);
	}
}

// handler for clicking the "Reset All Filters" button
static void reset_filters(GtkWidget *widget, gpointer data)
{
	for(int i = 0; i < 5; i++) {
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(spinner[i]), 0);
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(spinner[i + 5]), (i == 0 ? 0x100 : 0xFFFF));
	}
	gtk_tree_model_filter_refilter(GTK_TREE_MODEL_FILTER(sample_list_store_filtered));
}

// shims to make the spinners display their value in hex instead of decimal
static gint spinner_value_input(GtkSpinButton *spin, double *new_value, gpointer data)
{
	const gchar *text = gtk_entry_get_text(GTK_ENTRY(spin));
	long hex = strtol(text, NULL, 16);
	*new_value = hex;
	return TRUE;
}

static gboolean spinner_value_output(GtkSpinButton *spin, gpointer data)
{
	GtkAdjustment *adj = gtk_spin_button_get_adjustment(spin);
	int value = (int) gtk_adjustment_get_value(adj);
	gchar *text = g_strdup_printf("%X", value);
	gtk_entry_set_text(GTK_ENTRY(spin), text);
	g_free(text);
	return TRUE;
}

// handler for changing the values of any of the filters
static void filters_adjusted(GtkWidget *widget, gpointer data)
{
	GtkSpinButton *spin = GTK_SPIN_BUTTON(widget);
	gint index = GPOINTER_TO_INT(data);

	gint val = gtk_spin_button_get_value_as_int(spin);
	chart_filters[index] = val;

	gtk_tree_model_filter_refilter(GTK_TREE_MODEL_FILTER(sample_list_store_filtered));
}

// This overly complicated function copies the text from the clicked cell in a GtkTreeView
// to the clipboard, as if the user had selected it and pressed Ctrl+C.  This is required
// because GTK is silly and does not allow the user to select text from an individual column.
static void sample_list_click(GtkWidget *widget, GdkEventButton *event)
{
	GtkTreePath *path;
	GtkTreeModel *model;
	GtkTreeViewColumn *col;
	GtkTreeIter iter;
	gint x, y;
	gint *cell_x, *cell_y;
	GtkClipboard *clip;
	gchar *tempstr;
	int col_num;

	model = gtk_tree_view_get_model(GTK_TREE_VIEW(widget));
	path = gtk_tree_path_new();
	col = gtk_tree_view_column_new();

	cell_x = NULL;
	cell_y = NULL;

	x = event->x;
	y = event->y;

	gtk_tree_view_get_path_at_pos(GTK_TREE_VIEW(widget), x, y, &path, &col, cell_x, cell_y);

	// we have a GtkTreeViewColumn, but we need to know the numeric index for that column.
	col_num = -1;
	for(int i = 0; i < 5; i++) {
		if(col == sample_list_column[i]) {
			col_num = i;
			break;
		}
	}

	if(col_num > -1) {
		gtk_tree_model_get_iter(model, &iter, path);
		gtk_tree_model_get(model, &iter, col_num, &tempstr, -1);

		clip = gtk_clipboard_get_for_display(gdk_display_get_default(), GDK_SELECTION_CLIPBOARD);
		gtk_clipboard_set_text(clip, tempstr, -1);
		g_free(tempstr);
	}
}

// filter function for our sample table chart
static gboolean sample_table_filter(GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
	gboolean is_visible = TRUE;
	gchar *strings[5];
	long hex;
	int i;
	int *min_and_max = (int *)data;

	gtk_tree_model_get(model, iter, 0, &strings[0], 1, &strings[1], 2, &strings[2], 3, &strings[3], 4, &strings[4], -1);

	if(strings[0] == NULL) {
		is_visible = FALSE;
	} else {
		for(i = 0; i < 5; i++) {
			hex = strtol(strings[i], NULL, 16);

			if(hex < min_and_max[i] || hex > min_and_max[i + 5]) {
				is_visible = FALSE;
				break;
			}
		}
	}

	for(i = 0; i < 5; i++) {
		g_free(strings[i]);
	}

	return is_visible;
}

// sort function for our sample table chart.
static gint sample_table_sorter(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, gpointer data)
{
	gint sort_col = GPOINTER_TO_INT(data);
	gchar *val_a, *val_b;
	long hex_a, hex_b;

	gtk_tree_model_get(model, a, sort_col, &val_a, -1);
	gtk_tree_model_get(model, b, sort_col, &val_b, -1);

	if(val_a == NULL || val_b == NULL)
		return 0;

	hex_a = strtol(val_a, NULL, 16);
	hex_b = strtol(val_b, NULL, 16);
	g_free(val_a);
	g_free(val_b);

	if(hex_a < hex_b)
		return -1;
	else if(hex_a > hex_b)
		return 1;

	return 0;
}

static void play_spc_shim(GtkWidget *widget, gpointer data)
{
	GtkButton *button = GTK_BUTTON(widget);
	if(strcmp(gtk_button_get_label(button), "Play") == 0) {
		play_spc(spc);
		gtk_button_set_label(button, "Stop");
	} else {
		stop_spc();
		gtk_button_set_label(button, "Play");
	}
}

int main(int argc, char **argv)
{
	GtkWidget *window, *grid, *vbox, *sample_list, *scroll;
	GtkWidget *button_show_filters, *button_reset_filters, *grid_filters;
	GtkWidget *menubar, *menu, *menu_file, *menu_file_open, *menu_file_quit, *menu_help, *menu_help_about;
	GtkTreeModel *sample_list_store_sorted;
	GtkCellRenderer *renderer;
	GtkTreeSelection *sample_selection;
	GtkWidget *templabel;
	GtkAdjustment *adj;
	int i;

	gtk_init(&argc, &argv);

	// main window
	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(window), "Sample Table Tool");
	gtk_window_set_default_size(GTK_WINDOW(window), 600, 400);
	g_signal_connect(window, "destroy", G_CALLBACK(gtk_main_quit), NULL);

	vbox = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);
	gtk_container_add(GTK_CONTAINER(window), vbox);

	// menubar
	menubar = gtk_menu_bar_new();
	menu = gtk_menu_new();
	menu_file = gtk_menu_item_new_with_label("File");
	menu_file_open = gtk_menu_item_new_with_label("Open SPC");
	menu_file_quit = gtk_menu_item_new_with_label("Quit");
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(menu_file), menu);
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_file_open);
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_file_quit);
	gtk_menu_shell_append(GTK_MENU_SHELL(menubar), menu_file);
	menu = gtk_menu_new();
	menu_help = gtk_menu_item_new_with_label("Help");
	menu_help_about = gtk_menu_item_new_with_label("About");
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(menu_help), menu);
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_help_about);
	gtk_menu_shell_append(GTK_MENU_SHELL(menubar), menu_help);
	g_signal_connect_swapped(G_OBJECT(menu_file_open), "activate", G_CALLBACK(open_spc_dialog), window);
	g_signal_connect_swapped(G_OBJECT(menu_help_about), "activate", G_CALLBACK(about_dialog), window);
	g_signal_connect(G_OBJECT(menu_file_quit), "activate", G_CALLBACK(gtk_main_quit), NULL);
	gtk_box_pack_start(GTK_BOX(vbox), menubar, FALSE, FALSE, 3);

	// grid to display our ID666 info
	grid = gtk_grid_new();
	gtk_box_pack_start(GTK_BOX(vbox), grid, FALSE, FALSE, 3);

	add_label(&templabel, "Filename: ", grid, 0, 0, 1, 1, FALSE);
	add_label(&label_filename, "", grid, 1, 0, 1, 1, TRUE);
	add_label(&templabel, "Game Title: ", grid, 0, 1, 1, 1, FALSE);
	add_label(&label_game_title, "", grid, 1, 1, 1, 1, TRUE);
	add_label(&templabel, "Song Title: ", grid, 0, 2, 1, 1, FALSE);
	add_label(&label_song_title, "", grid, 1, 2, 1, 1, TRUE);
	add_label(&templabel, "Artist Title: ", grid, 0, 3, 1, 1, FALSE);
	add_label(&label_artist_name, "", grid, 1, 3, 1, 1, TRUE);
	add_label(&templabel, "Song Length: ", grid, 0, 4, 1, 1, FALSE);
	add_label(&label_song_length, "", grid, 1, 4, 1, 1, TRUE);

	button_play = gtk_button_new_with_label("Play");
	gtk_grid_attach(GTK_GRID(grid), button_play, 0, 5, 2, 1);
	g_signal_connect(G_OBJECT(button_play), "clicked", G_CALLBACK(play_spc_shim), NULL);

	// create a list view to handle our sample table (boy they make this complicated)
	sample_list_store = gtk_list_store_new(5, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);
	sample_list_store_filtered = gtk_tree_model_filter_new(GTK_TREE_MODEL(sample_list_store), NULL);
	// tell the list view how to filter the data
	gtk_tree_model_filter_set_visible_func(GTK_TREE_MODEL_FILTER(sample_list_store_filtered), sample_table_filter, chart_filters, NULL);
	sample_list_store_sorted = gtk_tree_model_sort_new_with_model(sample_list_store_filtered);

	// tell the list view how to sort the data
	for(i = 0; i < 5; i++) {
		gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(sample_list_store_sorted), i, sample_table_sorter, GINT_TO_POINTER(i), NULL);
	}
	gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(sample_list_store_sorted), 1, GTK_SORT_ASCENDING);
	sample_list = gtk_tree_view_new_with_model(GTK_TREE_MODEL(sample_list_store_sorted));

	// tell the list view how to render the data
	renderer = gtk_cell_renderer_text_new();
	char *chart_headings[5] = { "Sample #", "Start Address", "Loop Address", "Sample Length", "Loop Length" };
	for(i = 0; i < 5; i++) {
		sample_list_column[i] = gtk_tree_view_column_new_with_attributes(chart_headings[i], renderer, "text", i, NULL);
		gtk_tree_view_append_column(GTK_TREE_VIEW(sample_list), sample_list_column[i]);
		gtk_tree_view_column_set_sort_column_id(sample_list_column[i], i);
	}

	// allow clicking on a cell in the sample table to copy the contents to the clipboard
	sample_selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(sample_list));
	gtk_tree_selection_set_mode(sample_selection, GTK_SELECTION_NONE);
	g_signal_connect(G_OBJECT(sample_list), "button_press_event", G_CALLBACK(sample_list_click), NULL);
	gtk_widget_set_events(sample_list, GDK_BUTTON_PRESS_MASK);

	// collapsable display of Filters for the sample table
	button_show_filters = gtk_button_new_with_label("Show Filters");
	gtk_widget_set_halign(button_show_filters, GTK_ALIGN_END);
	gtk_widget_set_valign(button_show_filters, GTK_ALIGN_END);
	gtk_widget_set_vexpand(button_show_filters, FALSE);
	g_signal_connect(G_OBJECT(button_show_filters), "clicked", G_CALLBACK(filters_click), NULL);
	gtk_box_pack_start(GTK_BOX(vbox), button_show_filters, FALSE, FALSE, 3);
	frame_filters = gtk_frame_new("Filters");
	g_object_set(frame_filters, "halign", GTK_ALIGN_END, NULL);
	gtk_container_set_border_width(GTK_CONTAINER(frame_filters), 10);
	grid_filters = gtk_grid_new();
	gtk_container_set_border_width(GTK_CONTAINER(grid_filters), 5);
	add_label(&templabel, "Min", grid_filters, 1, 0, 1, 1, FALSE);
	add_label(&templabel, "Max", grid_filters, 2, 0, 1, 1, FALSE);
	for(i = 0; i < 5; i++) {
		add_label(&templabel, chart_headings[i], grid_filters, 0, i + 1, 1, 1, FALSE);

		adj = gtk_adjustment_new(chart_filters[i], 0, (i == 0 ? 0x100 : 0xFFFF), 1, 0x10, 0);
		spinner[i] = gtk_spin_button_new(adj, 1, 0);
		g_signal_connect(GTK_SPIN_BUTTON(spinner[i]), "input", G_CALLBACK(spinner_value_input), NULL);
		g_signal_connect(GTK_SPIN_BUTTON(spinner[i]), "output", G_CALLBACK(spinner_value_output), NULL);
		g_signal_connect(GTK_SPIN_BUTTON(spinner[i]), "value-changed", G_CALLBACK(filters_adjusted), GINT_TO_POINTER(i));
		gtk_grid_attach(GTK_GRID(grid_filters), spinner[i], 1, i + 1, 1, 1);

		adj = gtk_adjustment_new(chart_filters[i + 5], 0, (i == 0 ? 0x100 : 0xFFFF), 1, 0x10, 0);
		spinner[i + 5] = gtk_spin_button_new(adj, 1, 0);
		g_signal_connect(GTK_SPIN_BUTTON(spinner[i + 5]), "input", G_CALLBACK(spinner_value_input), NULL);
		g_signal_connect(GTK_SPIN_BUTTON(spinner[i + 5]), "output", G_CALLBACK(spinner_value_output), NULL);
		g_signal_connect(GTK_SPIN_BUTTON(spinner[i + 5]), "value-changed", G_CALLBACK(filters_adjusted), GINT_TO_POINTER(i + 5));
		gtk_grid_attach(GTK_GRID(grid_filters), spinner[i + 5], 2, i + 1, 1, 1);
	}
	button_reset_filters = gtk_button_new_with_label("Reset All Filters");
	//gtk_widget_set_vexpand(button_reset_filters, FALSE);
	//g_object_set(button_reset_filters, "halign", GTK_ALIGN_END, NULL);
	g_signal_connect(G_OBJECT(button_reset_filters), "clicked", G_CALLBACK(reset_filters), NULL);
	gtk_grid_attach(GTK_GRID(grid_filters), button_reset_filters, 0, 6, 1, 1);
	gtk_container_add(GTK_CONTAINER(frame_filters), grid_filters);
	gtk_box_pack_start(GTK_BOX(vbox), frame_filters, FALSE, FALSE, 3);

	// make the sample table scrollable if necessary
	scroll = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_min_content_height(GTK_SCROLLED_WINDOW(scroll), 400);
	gtk_container_add(GTK_CONTAINER(scroll), sample_list);
	gtk_box_pack_start(GTK_BOX(vbox), scroll, TRUE, TRUE, 3);

	gtk_widget_show_all(window);
	gtk_widget_hide(button_play); // hide the play button until an SPC is actually loaded
	gtk_widget_hide(frame_filters); // hide the filters grid until the "Filters" link is clicked

	gtk_main();

	return 0;
}
