/* SPC-related functions
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <pthread.h>
#include <ao/ao.h>
#include "spc.h"
#include "openspc.h"

#define BUFSIZE 32000

static pthread_t spc_thread;
static int thread_currently_exists = 0;
ao_device *audio_device = NULL;

struct spc_format *read_spc_from_file(char *filename)
{
	struct spc_format *spc = NULL;
	spc = malloc(sizeof(struct spc_format));
	if(spc == NULL) {
		perror("Unable to allocate memory for SPC file.\n");
		return NULL;
	}

	FILE *fp = fopen(filename, "rb");
	if(fp == NULL) {
		perror("Unable to open file.\n");
		return NULL;
	}

	int res = fread(spc, sizeof(struct spc_format), 1, fp);
	if(res < 1) {
		perror("Unable to read SPC file.\n");
		return NULL;
	}

	fclose(fp);

	if(strncmp(spc->header_constant, "SNES-SPC700 Sound File Data v", 29) != 0) {
		perror("This does not look like a valid SPC file.\n");
		return NULL;
	}

	return spc;
}

int calculate_sample_length(struct spc_format *spc, int sample_address)
{
	u8 *ptr = &spc->aram[sample_address];
	int len = 0;

	while(((*ptr) & 1) == 0) {
		len += 9;
		ptr += 9;
	}

	len += 9;
	return len;
}

void generate_sample_chart(struct spc_format *spc, int *out_num_samples, struct sample_chart ***out_chart)
{
	int index = spc->dsp_registers[0x5D] * 0x100;
	u16* sample_table = (u16*) &spc->aram[index];
	int i, num_samples;

	// count number of 'real' samples
	num_samples = 0;
	for(i = 0; i < 256 * 2; i += 2) {
		if(sample_table[i] != 0x0000 && sample_table[i] != 0xFFFF) {
			num_samples++;
		}
	}
	*out_num_samples = num_samples;

	// allocate memory
	struct sample_chart **chart;
	chart = calloc(num_samples, sizeof(struct sample_chart));
	if(chart == NULL) {
		perror("Unable to allocate memory for sample chart\n");
		return;
	}

	// build a nifty chart of the samples
	struct sample_chart *chart_entry = (struct sample_chart*) chart;
	for(i = 0; i < 256; i++) {
		if(sample_table[0] != 0x0000 && sample_table[0] != 0xFFFF) {
			chart_entry->sample_number = i + 1;
			chart_entry->sample_address = sample_table[0];
			chart_entry->loop_address = sample_table[1];
			chart_entry->sample_length = calculate_sample_length(spc, chart_entry->sample_address);
			chart_entry->loop_length = calculate_sample_length(spc, chart_entry->loop_address);
			chart_entry++;
		}
		sample_table += 2;
	}
	*out_chart = chart;
}

int sample_chart_comparator(const void *raw_a, const void *raw_b)
{
	struct sample_chart *a = (struct sample_chart*) raw_a;
	struct sample_chart *b = (struct sample_chart*) raw_b;

	if(a->sample_address < b->sample_address) {
		return -1;
	} else if(a->sample_address > b->sample_address) {
		return 1;
	} else {
		return 0;
	}
}

void dump_sample(struct spc_format *spc, u16 address, int length, char *filename)
{
	/*
	u8 *ptr = &spc->aram[address];
	u8 brr_block[9];
	int granularity, filter;
	u16 samp;

	FILE *fp = fopen(filename, "wb");
	if(fp == NULL) {
		perror("Unable to open file for writing");
		return;
	}

	for(int i = 0; i < length; i++) {
		memcpy(brr_block, ptr, 9);
		granularity = (brr_block[0] >> 4) & 0xF;
		filter = (brr_block[0] >> 2) & 0x3;

		// TODO

		ptr += 9;
	}

	fclose(fp);
	*/
}

static ao_device *init_audio_output(void)
{
	ao_device *device;
	ao_sample_format format;
	int default_driver;

	ao_initialize();
	default_driver = ao_default_driver_id();

	memset(&format, 0, sizeof(format));
	format.bits = 16;
	format.channels = 2;
	format.rate = 32000;
	format.byte_format = AO_FMT_LITTLE;

	device = ao_open_live(default_driver, &format, NULL);
	if (device == NULL) {
		printf("Error opening audio device.\n");
		return NULL;
	}

	return device;
}

/*
static void close_audio_output(void)
{
	ao_close(audio_device);
	ao_shutdown();
}
*/

static void spc_player_cleanup(void *arg)
{
	free(arg);
}

static void *spc_playing_thread(void *spc)
{
	int size;
	void *buf;

	buf = malloc(BUFSIZE);
	if(buf == NULL) {
		perror("Unable to allocate memory.");
		return NULL;
	}

	pthread_cleanup_push(spc_player_cleanup, buf);

	OSPC_Init(spc, sizeof(struct spc_format));

	do {
		pthread_testcancel();
		size = OSPC_Run(-1, buf, BUFSIZE);
		if(size > 0) {
			ao_play(audio_device, buf, size);
		}
	} while(size > 0);

	free(buf);

	pthread_cleanup_pop(0);

	return NULL;
}

void play_spc(struct spc_format *spc)
{
	// init AO if we haven't already
	if(audio_device == NULL) {
		audio_device = init_audio_output();
	}

	// if something is already playing, stop it first
	if(thread_currently_exists) {
		stop_spc();
	}

	// spin up a new thread to play this SPC
	if(pthread_create(&spc_thread, NULL, spc_playing_thread, spc)) {
		perror("Unable to create thread.");
		return;
	}
	thread_currently_exists = 1;
}

void stop_spc(void)
{
	pthread_cancel(spc_thread);

	pthread_join(spc_thread, NULL); // wait for thread to die
	thread_currently_exists = 0;
}
